import uuid
from typing import Optional, Union, List

from divinegift import main
from divinegift import logger
from divinegift.errors import ProducerNotSetError, ConsumerNotSetError

try:
    from confluent_kafka import Producer, Consumer, KafkaError, TopicPartition
    from confluent_kafka import OFFSET_STORED, OFFSET_BEGINNING, OFFSET_END
except ImportError:
    raise ImportError("confluent_kafka isn't installed. Run: pip install -U confluent_kafka")
try:
    from confluent_kafka import avro
    from confluent_kafka.avro import AvroProducer, AvroConsumer
    from confluent_kafka.avro.serializer import SerializerError
    from confluent_kafka.avro.error import ClientError
except ImportError:
    pass
try:
    from divinegift.string_avro import StringAvroConsumer
except NameError:
    pass


class KafkaConsumer:
    def __init__(self, logger_: Optional[logger.Logger] = None, avro_consumer=False, key_string=False, **configs):
        self.consumer: Union[Consumer, AvroConsumer, None] = None
        self.logger: Optional[logger.Logger] = None
        self.avro_consumer: bool = False

        try:
            self.init_consumer(logger_, avro_consumer, key_string, **configs)
        except Exception as ex:
            pass

    def init_consumer(self, logger_: Optional[logger.Logger] = None, avro_consumer=False, key_string=False, **configs):
        self.logger = logger_ if logger_ else logger.Logger()
        self.avro_consumer = avro_consumer
        if not self.avro_consumer:
            self.consumer = Consumer(configs)
        else:
            try:
                if key_string:
                    self.consumer = StringAvroConsumer(configs)
                else:
                    self.consumer = AvroConsumer(configs)
            except NameError:
                raise Exception("confluent_kafka[avro] isn't installed. Run: pip install -U confluent_kafka[avro]")
            except ClientError as ex:
                raise ClientError(str(ex))

    def close(self):
        if self.consumer is None:
            raise ConsumerNotSetError('Set consumer before!')
        self.consumer.close()
        self.consumer = None
        self.avro_consumer = False

    def subscribe(self, topics: Union[str, List], partition=None, offset=OFFSET_STORED):
        if self.consumer is None:
            raise ConsumerNotSetError('Set consumer before!')
        if partition is not None and offset != OFFSET_STORED:
            if isinstance(topics, List):
                for topic in topics:
                    self.set_offset(topic, partition, offset)
                    self.logger.log_info(f'[*] Consumer assigned to topic [{topic}] on partition [{partition}] and offset [{offset}]')
            else:
                self.set_offset(topics, partition, offset)
                self.logger.log_info(f'[*] Consumer assigned to topic [{topics}] on partition [{partition}] and offset [{offset}]')
        else:
            if isinstance(topics, str):
                topics = [topics]
            self.consumer.subscribe(topics)
            self.logger.log_info(f'[*] Consumer subscribed to topic {topics}')

    def set_offset(self, topic: str, partition: int, offset: int):
        if self.consumer is None:
            raise ConsumerNotSetError('Set consumer before!')
        topic_p = TopicPartition(topic, partition, offset)

        self.consumer.assign([topic_p])
        self.consumer.seek(topic_p)

    def consume_(self, num_messages=1, timeout=10):
        for msg_index in range(num_messages):
            msg = self.consumer.poll(timeout)
            if msg is None:
                continue
            if msg.error():
                self.logger.log_err(f"{'Avro' if self.avro_consumer else ''}Consumer error: {msg.error()}")
                continue
            self.logger.log_info(f'[<] Message received from {msg.topic()} ({msg.offset()} [{msg.partition()}])')
            self.logger.log_debug(f'[*] Received message: {msg.value()}')
            yield msg

    def consume(self, num_messages=1, timeout=10):
        for msg in self.consume_(num_messages=num_messages, timeout=timeout):
            yield msg

    def commit(self, message=None, *args, **kwargs):
        """
        .. py:function:: commit([message=None], [offsets=None], [asynchronous=True])

          Commit a message or a list of offsets.

          The ``message`` and ``offsets`` parameters are mutually exclusive. If neither is set, the current partition assignment's offsets are used instead. Use this method to commit offsets if you have 'enable.auto.commit' set to False.

          :param confluent_kafka.Message message: Commit the message's offset+1. Note: By convention, committed offsets reflect the next message to be consumed, **not** the last message consumed.
          :param list(TopicPartition) offsets: List of topic+partitions+offsets to commit.
          :param bool asynchronous: If true, asynchronously commit, returning None immediately. If False, the commit() call will block until the commit succeeds or fails and the committed offsets will be returned (on success). Note that specific partitions may have failed and the .err field of each partition should be checked for success.
        """
        if message:
            self.consumer.commit(message=message, *args, **kwargs)
        else:
            self.consumer.commit()

    def consume_txt(self, num_messages=1, timeout=10):
        for msg in self.consume_(num_messages=num_messages, timeout=timeout):
            if self.avro_consumer:
                yield msg.value()
            else:
                yield msg.value().decode('utf-8')


class KafkaProducer:
    def __init__(self, logger_: Optional[logger.Logger] = None, avro_producer=False, value_schema=None, key_schema=None,
                 **configs):
        self.producer: Union[Producer, AvroProducer, None] = None
        self.logger: Optional[logger.Logger] = None
        self.avro_producer: bool = False

        try:
            self.init_producer(logger_, avro_producer, value_schema, key_schema, **configs)
        except Exception as ex:
            pass

    def init_producer(self, logger_: Optional[logger.Logger] = None, avro_producer=False,
                      value_schema=None, key_schema=None, **configs):
        self.logger = logger_ if logger_ else logger.Logger()
        self.avro_producer = avro_producer
        if not self.avro_producer:
            self.producer = Producer(configs)
        else:
            try:
                if not key_schema:
                    key_schema = '{"type": "string"}'
                self.producer = AvroProducer(configs,
                                             default_value_schema=avro.loads(value_schema),
                                             default_key_schema=avro.loads(key_schema))
            except NameError:
                raise Exception("confluent_kafka isn't installed. Run: pip install -U confluent_kafka[avro]")
            except ClientError as ex:
                raise ClientError(str(ex))

    def close(self):
        if self.producer is None:
            raise ProducerNotSetError('Set producer before!')
        self.producer = None
        self.avro_producer = False

    def delivery_report(self, err, msg):
        """ Called once for each message produced to indicate delivery result.
            Triggered by poll() or flush(). """
        if err is not None:
            self.logger.log_info(f'Message delivery to {msg.topic()} failed: {err}')
            self.logger.log_debug(dir(msg))
            self.logger.log_debug(msg)
        else:
            self.logger.log_info(f'[>] Message delivered to {msg.topic()} [{msg.partition()}]')

    def produce(self, topic, message, key=None, partition=None, headers=None):
        if self.producer is None:
            raise ProducerNotSetError('Set producer before!')
        self.producer.poll(0)
        if not key:
            key = str(uuid.uuid4())
        if not self.avro_producer:
            key = key.encode('utf-8')
            if isinstance(message, str):
                message = message.encode('utf-8')
            else:
                json_obj = main.Json()
                json_obj.set_data(message)
                message = json_obj.dumps().encode('utf-8')

        if not partition:
            self.producer.produce(topic=topic, value=message, key=key,
                                  on_delivery=self.delivery_report, headers=headers)
        else:
            self.producer.produce(topic=topic, value=message, key=key,
                                  partition=partition, on_delivery=self.delivery_report, headers=headers)
        self.producer.flush()


class KafkaClient:
    def __init__(self, logger_: Optional[logger.Logger] = None):
        self.producer: Optional[KafkaProducer] = None
        self.avro_producer = False
        self.consumer: Optional[KafkaConsumer] = None
        self.avro_consumer = False
        self.logger = logger_ if logger_ else logger.Logger()

    def set_producer(self, **configs):
        self.producer = KafkaProducer(logger_=self.logger, avro_producer=False, **configs)

    def set_producer_avro(self, value_schema: str, **configs):
        self.producer = KafkaProducer(logger_=self.logger, avro_producer=True, value_schema=value_schema, **configs)

    def close_producer(self):
        self.producer.close()

    def set_consumer(self, **configs):
        self.consumer = KafkaConsumer(logger_=self.logger, avro_consumer=False, key_string=False, **configs)

    def set_consumer_avro(self, key_string=False, **configs):
        self.consumer = KafkaConsumer(logger_=self.logger, avro_consumer=True, key_string=key_string, **configs)

    def close_consumer(self):
        self.consumer.close()

    def send_message(self, topic, messages):
        if isinstance(messages, list):
            for message in messages:
                self.producer.produce(topic, message)
        else:
            self.producer.produce(topic, messages)

    def read_messages(self, topic, partition=None, offset=OFFSET_STORED):
        self.consumer.subscribe(topic, partition, offset)
        while True:
            for msg in self.consumer.consume(num_messages=1, timeout=10):
                if msg is None:
                    continue
                yield msg


if __name__ == '__main__':
    pass



name = "DivineGift"

version = '2.15.16'

__all__ = [
    "main",
    "logger",
    "banner",
    "config",
    "mailer",
    "sender",
    "db",
    "translit",
    "excel",
    "templator",
    "kafka_client",
    "cipher",
    "zabbix_agent",
    "kthreads",
    "app",
    "aims",
    "timer",
    "smb"
]



class SenderNotInitializedError(Exception):
    pass


class KillException(Exception):
    pass


class ProducerNotSetError(Exception):
    pass


class ConsumerNotSetError(Exception):
    pass


class MethodNotAllowedError(Exception):
    pass


class EmptyConfigError(Exception):
    pass


class ConfigNotFoundError(Exception):
    pass


class DBNotConnectedError(Exception):
    pass


class NoFileNamePassedError(Exception):
    pass



from deprecation import deprecated

from divinegift import main, version

import os
import logging
from logging.handlers import TimedRotatingFileHandler

from typing import List, Dict, Optional


class Logger:
    def __init__(self):
        self.logger: Optional[logging.Logger] = None
        self.log_dir = None
        self.log_name = None
        self.formatter = logging.Formatter('%(levelname)-8s [%(asctime)s] %(message)s')

    def set_log_level_(self):
        self.set_log_level(**main.get_log_param(main.get_args()))

    def set_log_level(self, log_level: str, log_name: str = None, log_dir: str = './logs/', when: str = 'midnight',
                      interval: int = 1, backup_count: int = 7):
        """
            This set up log_level and name of logfile
            :param log_level: String with log_level (e.g. 'INFO')
            :param log_name: Name of file with logs
            :param log_dir: Directory which should keep logs
            :param when: When rotate log
            :param interval: How often
            :param backup_count: How many versions to keep
            :return: None
        """

        self.logger = logging.getLogger('Rotating')
        # Remove all existing handlers
        for handler in self.logger.handlers[:]:
            self.logger.removeHandler(handler)
        numeric_level = getattr(logging, log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % log_level)

        self.logger.setLevel(numeric_level)

        if log_name:
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            handler = TimedRotatingFileHandler(os.path.join(log_dir, log_name), when=when, interval=interval,
                                               backupCount=backup_count, encoding='utf-8')
            self.log_name = log_name
            self.log_dir = log_dir
        else:
            handler = logging.StreamHandler()

        handler.setFormatter(self.formatter)
        self.logger.addHandler(handler)

    def log_debug(self, *args, separator: str = ' '):
        """
        Logging a debug message
        :return: None
        """
        if not self.logger:
            self.set_log_level_()
        self.logger.debug(separator.join([str(x) for x in args]))

    def log_info(self, *args, separator: str = ' '):
        """
        Logging a info
        :param separator: Separator
        :return: None
        """
        if not self.logger:
            self.set_log_level_()
        self.logger.info(separator.join([str(x) for x in args]))

    def log_warn(self, *args, separator: str = ' '):
        """
        Logging a warning
        :param separator: Separator
        :return: None
        """
        if not self.logger:
            self.set_log_level_()
        self.logger.warning(separator.join([str(x) for x in args]))

    def log_err(self, *args, separator: str = ' '):
        """
        Logging an error with monitoring if parameters were filled
        :param separator: Separator
        :return: None
        """
        if not self.logger:
            self.set_log_level_()
        self.logger.exception(separator.join([str(x) for x in args]))

    def log_crit(self, *args, separator: str = ' '):
        if not self.logger:
            self.set_log_level_()
        self.logger.critical(separator.join([str(x) for x in args]))


@deprecated(deprecated_in='2.8.0', current_version=version, details='Use class Logger instead')
def log_debug(*args, separator: str = ' '):
    get_logger().log_debug(*args, separator=separator)


@deprecated(deprecated_in='2.8.0', current_version=version, details='Use class Logger instead')
def log_info(*args, separator: str = ' '):
    get_logger().log_info(*args, separator=separator)


@deprecated(deprecated_in='2.8.0', current_version=version, details='Use class Logger instead')
def log_warning(*args, separator: str = ' '):
    get_logger().log_warn(*args, separator=separator)


@deprecated(deprecated_in='2.8.0', current_version=version, details='Use class Logger instead')
def log_err(*args, separator: str = ' ', src: str = None, mode: List = None, channel: Dict = None):
    get_logger().log_err(*args, separator=separator)


@deprecated(deprecated_in='2.8.0', current_version=version, details='Use class Logger instead')
def log_crit(*args, separator: str = ' '):
    get_logger().log_crit(*args, separator=separator)


def set_loglevel(log_level: str, log_name: str = None, log_dir: str = './logs/',
                 when: str = 'midnight', interval: int = 1, backupCount: int = 7):
    get_logger().set_log_level(log_level, log_name, log_dir, when, interval, backupCount)


logger_ = Logger()


def get_logger():
    return logger_


if __name__ == '__main__':
    pass



from typing import Union, Optional

from cryptography.fernet import InvalidToken

from divinegift import main
from divinegift import logger
from divinegift import cipher
#########################################################################
import sys
import os
from datetime import datetime
from copy import deepcopy

from divinegift.errors import EmptyConfigError, ConfigNotFoundError


class Settings:
    parse_settings_cnt = 0

    def __init__(self, file_: Optional[str] = None, ck_fname: Optional[str] = 'key.ck', encoding: str = 'utf-8',
                 log_changes: bool = False, use_yaml: bool = True, ignore_parse_cnt: bool = True,
                 logger_: logger.Logger = None):
        self.settings: dict = {}
        self.settings_encrypted: dict = {}
        self.encrypted_values: dict = {}
        self.cipher_key = None
        self.cipher_key_fname = None
        self.passwords_cnt = 0
        self.logger = logger_ if logger_ else logger.Logger()

        self.file_extensions_allowed = ['json', 'yaml', 'yml', 'ini', 'cfg', 'conf']

        if file_:
            self.parse_settings(file_, ck_fname, encoding, log_changes, use_yaml, ignore_parse_cnt)

    def get_settings(self, param: str = None, default: Union[str, list, dict, bool] = None):
        if param:
            return self.settings.get(param, default)
        return self.settings

    def set_settings(self, json_str: dict):
        self.settings = deepcopy(json_str)
        if self.cipher_key_fname:
            self.decrypt_passwords()
        self.settings_encrypted = deepcopy(json_str)

    def decrypt_passwords(self):
        self.logger.log_debug('Try to decrypt all passwords')
        for conn_name in self.settings.keys():
            try:
                for pass_field in self.settings.get(conn_name).keys():
                    try:
                        self.decrypt_password(conn_name, pass_field)
                        self.logger.log_debug(f'Password for {conn_name}[{pass_field}] decrypted')
                        self.passwords_cnt += 1
                        self.encrypted_values[conn_name] = pass_field
                    except:
                        pass
            except:
                pass

    def find_config_file(self, file):
        d, f = os.path.split(file)
        for ext in self.file_extensions_allowed:
            f_n = f.replace(main.get_file_extension(f), '') + f'.{ext}'
            if os.path.exists(os.path.join(d, f_n)):
                self.logger.log_info(f'Found config file with extension: {f_n}')
                return os.path.join(d, f_n)
        raise ConfigNotFoundError(f'There is no config file in directory {d} with allowed extensions: [{self.file_extensions_allowed}]')

    def parse_settings(self, file_: str = './settings.yaml', ck_fname: str = 'key.ck', encoding: str = 'utf-8',
                       log_changes: bool = False, use_yaml: bool = True, ignore_parse_cnt: bool = True):
        """
        Parse settings from file_
        :param file_: Filename with settings
        :param ck_fname: Filename with key.ck
        :param encoding: Encoding
        :param log_changes: Set to False for not showing logs of parsing configs
        :param use_yaml: True by default. Set to False if settings in JSON
        :param ignore_parse_cnt: Set to False if repeatedly reads configs
        :return: None
        """
        if not ignore_parse_cnt:
            self.__class__.parse_settings_cnt += 1
            if self.__class__.parse_settings_cnt > 2:
                raise EmptyConfigError('Settings not parsed. Is there any data?')

        if not os.path.exists(file_):
            file_ = self.find_config_file(file_)
        
        if use_yaml:
            c = main.Yaml(file_, encoding=encoding)
        else:
            c = main.Json(file_, encoding=encoding)

        json_data = c.parse()

        if log_changes and json_data:
            dict_c = main.dict_compare(self.settings_encrypted, json_data)
            added, removed, modified, same = dict_c.values()
            if len(added) > 0:
                for r in list(added):
                    self.logger.log_warn(f'Added {r}: {json_data.get(r)}')
            if len(removed) > 0:
                for r in list(removed):
                    self.logger.log_warn(f'Removed {r}: {self.settings_encrypted.get(r)}')
            if len(modified) > 0:
                for r in list(modified):
                    self.logger.log_warn(f'Modified {r}: {modified.get(r)[0]} -> {modified.get(r)[1]}')
        elif not json_data:
            try:
                # self.convert_config(file_, file_, encoding=encoding, use_yaml=use_yaml)
                self.parse_settings(file_, encoding, ck_fname, not use_yaml)
            except EmptyConfigError as ms:
                self.logger.log_err(ms)
                sys.exit(-1)
            except Exception as ex:
                raise Exception('Settings not parsed. Is there any data 2?')

        self.cipher_key_fname = ck_fname

        self.set_settings(json_data)

    def initialize_cipher(self, ck_fname: str = 'key.ck'):
        self.cipher_key = cipher.get_key()
        cipher.write_key(ck_fname, self.cipher_key)

    def encrypt_password(self, encrypt_field: str, encrypt_value: str = 'db_pass'):
        if not self.cipher_key:
            self.cipher_key = cipher.read_key(self.cipher_key_fname)
            if not self.cipher_key:
                self.initialize_cipher()
        cipher_ = cipher.get_cipher(self.cipher_key)
        self.settings[encrypt_field][encrypt_value] = cipher.encrypt_str(self.settings[encrypt_field][encrypt_value], cipher_)

    def decrypt_password(self, encrypt_field: str, encrypt_value: str = 'db_pass'):
        if not self.cipher_key:
            self.cipher_key = cipher.read_key(self.cipher_key_fname)
        cipher_ = cipher.get_cipher(self.cipher_key)
        self.settings[encrypt_field][encrypt_value] = cipher.decrypt_str(self.settings[encrypt_field][encrypt_value], cipher_)

    def encode_password(self, encrypt_field: str, encrypt_value: str = 'db_pass'):
        settings = self.get_settings().copy()
        settings[encrypt_field][encrypt_value] = cipher.encode_password(settings[encrypt_field][encrypt_value])
        self.set_settings(settings)

    def decode_password(self, encrypt_field: str, encrypt_value: str = 'db_pass'):
        settings = self.get_settings().copy()
        settings[encrypt_field][encrypt_value] = cipher.decode_password(settings[encrypt_field][encrypt_value])
        self.set_settings(settings)

    def save_settings(self, file_: str = './settings.yml', encoding='utf-8', use_yaml=True):
        if use_yaml:
            c = main.Yaml(file_, encoding=encoding)
        else:
            c = main.Json(file_, encoding=encoding)
        c.set_data(eval(str(self.get_settings())))
        c.create()

    def check_config_still_encrypted(self):
        for encrypted_field, encrypted_value in self.encrypted_values.items():
            try:
                cipher_ = cipher.get_cipher(self.cipher_key)
                cipher.decrypt_str(self.settings[encrypted_field][encrypted_value], cipher_)
            except InvalidToken:
                self.encrypt_password(encrypted_field, encrypted_value)

    @staticmethod
    def convert_config(file_from: str = './settings.yml', file_to: str = './settings.yml',
                       encoding='utf-8', use_yaml=True):
        if use_yaml:
            j = main.Json(file_from, encoding=encoding)
            y = main.Yaml(file_to, encoding=encoding)
            j.parse()
            y.set_data(j.get_data())
            y.create()
        else:
            y = main.Yaml(file_from, encoding=encoding)
            j = main.Json(file_to, encoding=encoding)
            y.parse()
            j.set_data(y.get_data())
            j.create()

    @staticmethod
    def get_version(fname=None):
        if not fname:
            fname = sys.argv[0]
        filename = main.get_list_files(os.path.dirname(fname), filter=os.path.basename(fname), add_path=True)[-1]
        version = datetime.fromtimestamp(os.path.getmtime(filename))
        return datetime.strftime(version, '%y%m%d')


if __name__ == '__main__':
    pass
